"""
application independent commands
"""

import logging

import urwid

from sen.exceptions import NotifyError
from sen.tui.buffer import HelpBuffer, TreeBuffer
from sen.tui.commands.base import (
    register_command, SameThreadCommand,
    Option, Argument,
    NoSuchCommand
)
from sen.util import log_traceback, log_last_traceback


logger = logging.getLogger(__name__)


class LogTracebackMixin:
    @log_traceback
    def do(self, fn, *args, **kwargs):
        try:
            fn(*args, **kwargs)
        except NotifyError as ex:
            logger.error(repr(ex))
            self.ui.notify_message(str(ex), level="error")
        except Exception as ex:
            logger.error(repr(ex))
            self.ui.notify_message("Command failed: %s" % str(ex), level="error")
            raise


@register_command
class QuitCommand(SameThreadCommand):
    name = "quit"
    # TODO: make this configurable by asking whether to quit or not
    description = "Quit sen. No questions asked."

    def run(self):
        self.ui.quit()


@register_command
class KillBufferCommand(SameThreadCommand):
    name = "kill-buffer"
    description = "Remove currently displayed buffer."
    options_definitions = [
        Option("quit-if-no-buffer", "Quit when there's no buffer left", default=False)
    ]

    def run(self):
        buffers_left = self.ui.remove_current_buffer(close_if_no_buffer=self.arguments.quit_if_no_buffer)
        if buffers_left is None:
            self.ui.notify_message("Last buffer will not be removed.")
        elif buffers_left == 0:
            self.ui.run_command(QuitCommand.name)


@register_command
class SelectBufferCommand(SameThreadCommand):
    name = "select-buffer"
    description = "Display buffer with selected index."
    arguments_definitions = [
        Argument("index", "Index of buffer to display", default=1, action=int)
    ]

    def run(self):
        self.ui.pick_and_display_buffer(self.arguments.index)


@register_command
class SelectNextBufferCommand(SelectBufferCommand):
    name = "select-next-buffer"
    description = "Display next buffer."

    def run(self):
        self.arguments.set_argument("index", self.ui.current_buffer_index + 1)
        super().run()


@register_command
class SelectPreviousBufferCommand(SelectBufferCommand):
    name = "select-previous-buffer"
    description = "Display previous buffer."

    def run(self):
        self.arguments.set_argument("index", self.ui.current_buffer_index - 1)
        super().run()


@register_command
class DisplayBufferCommand(SameThreadCommand):
    name = "display-buffer"  # TODO: make this a universal display function

    arguments_definitions = [Argument("buffer", "Buffer instance to show.")]
    description = "This is an internal command and doesn't work from command line."

    def run(self):
        # TODO: doesn't work!, the method expects buffer class, not string
        self.ui.add_and_display_buffer(self.arguments.buffer)


@register_command
class DisplayHelpCommand(SameThreadCommand):
    name = "help"
    description = "Display help about buffer or command. When 'query' is not specified " + \
        "help for current buffer is being displayed."
    arguments_definitions = [Argument("query", "input string: command, buffer name")]

    def run(self):
        if self.arguments.query is None:
            self.ui.add_and_display_buffer(HelpBuffer(self.ui, self.buffer))
            return
        try:
            command = self.ui.commander.get_command(self.arguments.query)
        except NoSuchCommand:
            self.ui.notify_message("There is no such command: %r" % self.arguments.query)
        else:
            self.ui.add_and_display_buffer(HelpBuffer(self.ui, command))
            return
        # TODO: help view for commands could be displayed in footer


@register_command
class DisplayLayersCommand(DisplayBufferCommand):
    name = "layers"
    description = "open a tree view of all image layers (`docker images --tree` equivalent)"

    def run(self):
        self.arguments.set_argument("buffer", TreeBuffer(self.ui, self.docker_backend))
        super().run()


@log_traceback
def run_command_callback(ui, docker_object, edit_widget, text_input):
    logger.debug("%r %r", edit_widget, text_input)
    if "\n" in text_input:
        inp = text_input.strip()
        inp = inp.replace("\n", "")
        # first restore statusbar and then run the command
        ui.prompt_bar = None
        ui.set_focus("body")
        try:
            ui.run_command(inp, docker_object=docker_object)
        except NoSuchCommand as ex:
            logger.info("non-existent command initiated: %r", inp)
            ui.notify_message(str(ex), level="error")
        except Exception as ex:
            logger.info("command %r failed: %r", inp, ex)
            ui.notify_message("Error while running command '{}': {}".format(
                inp, ex
            ), level="error")
            log_last_traceback()
        ui.reload_footer()


@register_command
class PromptCommand(SameThreadCommand):
    name = "prompt"
    description = "Customize and pre-populate prompt with initial data."
    options_definitions = [
        Option("prompt-text", "Text forming the actual prompt", default=":"),
        Option("initial-text", "Prepopulated text", default="")
    ]

    def run(self):
        """
        prompt for text input.
        """
        # set up widgets
        leftpart = urwid.Text(self.arguments.prompt_text, align='left')
        editpart = urwid.Edit(multiline=True, edit_text=self.arguments.initial_text)

        # build promptwidget
        edit = urwid.Columns([
            ('fixed', len(self.arguments.prompt_text), leftpart),
            ('weight', 1, editpart),
        ])
        self.ui.prompt_bar = urwid.AttrMap(edit, "main_list_dg")

        self.ui.reload_footer()
        self.ui.set_focus("footer")

        urwid.connect_signal(editpart, "change", run_command_callback,
                             user_args=[self.ui, self.docker_object])


@register_command
class SearchCommand(SameThreadCommand, LogTracebackMixin):
    name = "search"
    description = "search and highlight (provide empty string to disable searching)"
    arguments_definitions = [
        Argument("query", "Input string to search for")
    ]
    aliases = ["/"]

    def run(self):
        # TODO: implement incsearch
        #   - match needs to be highlighted somehow, not with focus though
        #     - a line could split into a Text with multiple markups
        query = self.arguments.query if self.arguments.query is not None else ""
        self.do(self.ui.current_buffer.find_next, query)


@register_command
class SearchNextCommand(SameThreadCommand, LogTracebackMixin):
    name = "search-next"
    description = "next search occurrence"

    def run(self):
        self.do(self.ui.current_buffer.find_next)


@register_command
class SearchPreviousCommand(SameThreadCommand, LogTracebackMixin):
    name = "search-previous"
    description = "previous search occurrence"

    def run(self):
        self.do(self.ui.current_buffer.find_previous)


@register_command
class SearchCommand(SameThreadCommand, LogTracebackMixin):
    name = "filter"
    description = """\
Display only lines matching provided query (provide empty query to clear filtering)

In main listing, you may specify more precise query with these space-separated filters:
* t[ype]=c[ontainer[s]]
* t[ype]=i[mage[s]]
* s[tate]=r[unning])

Examples
* "type=container" - show only containers (short equivalent is "t=c")
* "type=image fedora" - show images with string "fedora" in name (equivalent "t=i fedora")\
"""

    arguments_definitions = [
        Argument("query", "Input query string", default="")
    ]

    def run(self):
        # TODO: realtime list change would be mindblowing
        self.do(self.ui.current_buffer.filter, self.arguments.query)


@register_command
class RefreshCurrentBufferCommand(SameThreadCommand):
    name = "refresh"
    description = "Refresh current buffer."

    def run(self):
        self.ui.current_buffer.refresh()


@register_command
class SearchPreviousCommand(SameThreadCommand):
    name = "toggle-live-updates"
    description = "Toggle realtime updates of the interface (this is useful when you are " + \
                  "removing multiple objects and don't want the listing change during " + \
                  "the process so you accidentally remove something)."

    def run(self):
        self.ui.current_buffer.widget.toggle_realtime_events()


@register_command
class RedrawUI(SameThreadCommand):
    name = "redraw"
    description = "Redraw user interface."

    def run(self):
        self.ui.loop.screen.clear()
        self.ui.refresh()



from sen.net import NetData

empty_inspect = {
    "NetworkSettings": {
        "SecondaryIPAddresses": None,
        "EndpointID": "",
        "IPv6Gateway": "",
        "IPAddress": "",
        "LinkLocalIPv6PrefixLen": 0,
        "Networks": None,
        "GlobalIPv6PrefixLen": 0,
        "MacAddress": "",
        "GlobalIPv6Address": "",
        "SandboxID": "",
        "SecondaryIPv6Addresses": None,
        "SandboxKey": "",
        "HairpinMode": False,
        "IPPrefixLen": 0,
        "Bridge": "",
        "Gateway": "",
        "Ports": None,
        "LinkLocalIPv6Address": ""
    },
}
populated_inspect = {
    "NetworkSettings": {
        "HairpinMode": False,
        "SandboxKey": "/var/run/docker/netns/7c1ad2c7d916",
        "MacAddress": "02:42:ac:11:00:07",
        "SandboxID": "7c1ad2c7d916cdbb6f92ef22feb420b67b186eb719d7b968fa508820e55617ab",
        "SecondaryIPAddresses": None,
        "Ports": {
            "8080/tcp": [
                {
                    "HostPort": "31003",
                    "HostIp": "0.0.0.0"
                }
            ]
        },
        "SecondaryIPv6Addresses": None,
        "IPAddress": "172.17.0.7",
        "GlobalIPv6Address": "",
        "EndpointID": "4ec3605fe3c70ef85cacee52d29db70c67f87cf85fd8021bf9e35ae09d4a5be2",
        "LinkLocalIPv6Address": "",
        "Networks": {
            "bridge": {
                "MacAddress": "02:42:ac:11:00:07",
                "Links": None,
                "IPAMConfig": None,
                "IPAddress": "172.17.0.7",
                "NetworkID": "2e90eba6e82abb1fe2ba82bfafbb1cf675e92cb9b33e667c78d5ab12cc3ab5b5",
                "GlobalIPv6Address": "",
                "GlobalIPv6PrefixLen": 0,
                "Aliases": None,
                "IPv6Gateway": "",
                "Gateway": "172.17.0.1",
                "EndpointID": "4ec3605fe3c70ef85cacee52d29db70c67f87cf85fd8021bf9e35ae09d4a5be2",
                "IPPrefixLen": 16
            }
        },
        "Bridge": "",
        "LinkLocalIPv6PrefixLen": 0,
        "IPv6Gateway": "",
        "Gateway": "172.17.0.1",
        "GlobalIPv6PrefixLen": 0,
        "IPPrefixLen": 16
    },
    "Config": {
        "ExposedPorts": {
            "8787/tcp": {}
        },
    }
}


def test_empty_network():
    nd = NetData(empty_inspect)
    assert nd.ips == {}
    assert nd.ports == {}


def test_populated_network():
    nd = NetData(populated_inspect)
    assert nd.ips == {"bridge": {"ip_address4": "172.17.0.7"},
                      "default": {"ip_address4": "172.17.0.7"}}
    assert nd.ports == {"8080": "31003", "8787": None}



import os
import logging
import functools
import traceback
import time
from datetime import datetime

from docker.errors import APIError

from sen.constants import PROJECT_NAME, LOG_FILE_NAME


logger = logging.getLogger(__name__)


def _ensure_unicode(s):
    try:
        return s.decode("utf-8")
    except AttributeError:
        return s


def log_last_traceback():
    logger.error(traceback.format_exc())


def log_traceback(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger.info("function %s is about to be started", func)
        try:
            response = func(*args, **kwargs)
        except Exception:
            log_last_traceback()
        else:
            logger.info("function %s finished", func)
            # TODO: how long it took?
            return response
    return wrapper


def setup_dirs():
    """Make required directories to hold logfile.

    :returns: str
    """
    try:
        top_dir = os.path.abspath(os.path.expanduser(os.environ["XDG_CACHE_HOME"]))
    except KeyError:
        top_dir = os.path.abspath(os.path.expanduser("~/.cache"))
    our_cache_dir = os.path.join(top_dir, PROJECT_NAME)
    os.makedirs(our_cache_dir, mode=0o775, exist_ok=True)
    return our_cache_dir


def get_log_file_path():
    return os.path.join(setup_dirs(), LOG_FILE_NAME)


def humanize_bytes(bytesize, precision=2):
    """
    Humanize byte size figures

    https://gist.github.com/moird/3684595
    """
    abbrevs = (
        (1 << 50, 'PB'),
        (1 << 40, 'TB'),
        (1 << 30, 'GB'),
        (1 << 20, 'MB'),
        (1 << 10, 'kB'),
        (1, 'bytes')
    )
    if bytesize == 1:
        return '1 byte'
    for factor, suffix in abbrevs:
        if bytesize >= factor:
            break
    if factor == 1:
        precision = 0
    return '%.*f %s' % (precision, bytesize / float(factor), suffix)


def humanize_time(value):
    abbrevs = (
        (1, "now"),
        (2, "{seconds} seconds ago"),
        (59, "{seconds} seconds ago"),
        (60, "{minutes} minute ago"),
        (119, "{minutes} minute ago"),
        (120, "{minutes} minutes ago"),
        (3599, "{minutes} minutes ago"),
        (3600, "{hours} hour ago"),
        (7199, "{hours} hour ago"),
        (86399, "{hours} hours ago"),
        (86400, "{days} day ago"),
        (172799, "{days} day ago"),
        (172800, "{days} days ago"),
        (172800, "{days} days ago"),
        (2591999, "{days} days ago"),
        (2592000, "{months} month ago"),
        (5183999, "{months} month ago"),
        (5184000, "{months} months ago"),
    )
    n = datetime.now()
    delta = n - value
    for guard, message in abbrevs:
        s = int(delta.total_seconds())
        if guard >= s:
            break
    return message.format(seconds=delta.seconds, minutes=int(delta.seconds // 60),
                          hours=int(delta.seconds // 3600), days=delta.days,
                          months=int(delta.days // 30))


# # This function is able to crash python b/c it may write monster-amount of data.
# #   Use it only for debugging, do not ship it!
# def log_vars_from_tback(process_frames=5):
#     for th in threading.enumerate():
#         try:
#             thread_frames = sys._current_frames()[th.ident]
#         except KeyError:
#             continue
#         logger.debug(''.join(traceback.format_stack(thread_frames)))
#
#     logger.error(traceback.format_exc())
#     if process_frames <= 0:
#         return
#     tb = sys.exc_info()[2]
#     while 1:
#         if not tb.tb_next:
#             break
#         tb = tb.tb_next
#     stack = []
#     f = tb.tb_frame
#     while f:
#         stack.append(f)
#         f = f.f_back
#     for frame in stack[:process_frames]:
#         logger.debug("frame %s:%s", frame.f_code.co_filename, frame.f_lineno)
#         for key, value in frame.f_locals.items():
#             try:
#                 logger.debug("%20s = %s", key, value)
#             except Exception:
#                 logger.debug("%20s = CANNOT PRINT VALUE", key)
#
#             # self_instance = frame.f_locals.get("self", None)
#             # if not self_instance:
#             #     continue
#             # for key in dir(self_instance):
#             #     if key.startswith("__"):
#             #         continue
#             #     try:
#             #         value = getattr(self_instance, key, None)
#             #         logger.debug("%20s = %s", "self." + key, value)
#             #     except Exception:
#             #         logger.debug("%20s = CANNOT PRINT VALUE", "self." + key)


# this is taken directly from docker client:
#   https://github.com/docker/docker/blob/28a7577a029780e4533faf3d057ec9f6c7a10948/api/client/stats.go#L309
def calculate_cpu_percent(d):
    cpu_count = len(d["cpu_stats"]["cpu_usage"]["percpu_usage"])
    cpu_percent = 0.0
    cpu_delta = float(d["cpu_stats"]["cpu_usage"]["total_usage"]) - \
                float(d["precpu_stats"]["cpu_usage"]["total_usage"])
    system_delta = float(d["cpu_stats"]["system_cpu_usage"]) - \
                   float(d["precpu_stats"]["system_cpu_usage"])
    if system_delta > 0.0:
        cpu_percent = cpu_delta / system_delta * 100.0 * cpu_count
    return cpu_percent

# again taken directly from docker:
#   https://github.com/docker/cli/blob/2bfac7fcdafeafbd2f450abb6d1bb3106e4f3ccb/cli/command/container/stats_helpers.go#L168
# precpu_stats in 1.13+ is completely broken, doesn't contain any values
def calculate_cpu_percent2(d, previous_cpu, previous_system):
    # import json
    # du = json.dumps(d, indent=2)
    # logger.debug("XXX: %s", du)
    cpu_percent = 0.0
    cpu_total = float(d["cpu_stats"]["cpu_usage"]["total_usage"])
    cpu_delta = cpu_total - previous_cpu
    cpu_system = float(d["cpu_stats"]["system_cpu_usage"])
    system_delta = cpu_system - previous_system
    online_cpus = d["cpu_stats"].get("online_cpus", len(d["cpu_stats"]["cpu_usage"]["percpu_usage"]))
    if system_delta > 0.0:
        cpu_percent = (cpu_delta / system_delta) * online_cpus * 100.0
    return cpu_percent, cpu_system, cpu_total


def calculate_blkio_bytes(d):
    """

    :param d:
    :return: (read_bytes, wrote_bytes), ints
    """
    bytes_stats = graceful_chain_get(d, "blkio_stats", "io_service_bytes_recursive")
    if not bytes_stats:
        return 0, 0
    r = 0
    w = 0
    for s in bytes_stats:
        if s["op"] == "Read":
            r += s["value"]
        elif s["op"] == "Write":
            w += s["value"]
    return r, w


def calculate_network_bytes(d):
    """

    :param d:
    :return: (received_bytes, transceived_bytes), ints
    """
    networks = graceful_chain_get(d, "networks")
    if not networks:
        return 0, 0
    r = 0
    t = 0
    for if_name, data in networks.items():
        logger.debug("getting stats for interface %r", if_name)
        r += data["rx_bytes"]
        t += data["tx_bytes"]
    return r, t


def graceful_chain_get(d, *args, default=None):
    t = d
    for a in args:
        try:
            t = t[a]
        except (KeyError, ValueError, TypeError, AttributeError):
            logger.debug("can't get %r from %s", a, t)
            return default
    return t


def repeater(call, args=None, kwargs=None, retries=4):
    """
    repeat call x-times: docker API is just awesome

    :param call: function
    :param args: tuple, args for function
    :param kwargs: dict, kwargs for function
    :param retries: int, how many times we try?
    :return: response of the call
    """
    args = args or ()
    kwargs = kwargs or {}
    t = 1.0
    for x in range(retries):
        try:
            return call(*args, **kwargs)
        except APIError as ex:
            logger.error("query #%d: docker returned an error: %r", x, ex)
        except Exception as ex:
            # this may be pretty bad
            log_last_traceback()
            logger.error("query #%d: generic error: %r", x, ex)
        t *= 2
        time.sleep(t)


class OrderedSet(list):
    def append(self, p_object):
        if p_object in self:
            self.remove(p_object)
        return super().append(p_object)




#!/usr/bin/env python3

import os

from setuptools import setup, find_packages

# vcs+proto://host/path@revision#egg=project-version
# this is latest upstream commit (April 2017)
# upstream maintainer of urwidtrees doesn't maintain PyPI urwidtrees
urwidtrees_source = "git+https://github.com/pazz/urwidtrees.git@9142c59d3e41421ff6230708d08b6a134e0a8eed#egg=urwidtrees-1.0.3.dev"

requirements = [
    "urwid",
    "docker",
    "urwidtrees"
]

setup(
    name='sen',
    version='0.6.2',
    description="Terminal User Interface for Docker Engine",
    author='Tomas Tomecek',
    author_email='tomas@tomecek.net',
    url="https://github.com/TomasTomecek/sen/",
    license="MIT",
    entry_points={
        'console_scripts': ['sen=sen.cli:main'],
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=requirements,
    dependency_links=[urwidtrees_source],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Intended Audience :: Developers',
                 'Intended Audience :: System Administrators',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: POSIX :: Linux',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3 :: Only',
                 'Programming Language :: Python :: 3.4',
                 'Topic :: System :: Monitoring',
                 ],
)
